import math


# Helper class to calculate PnL and returns over a single instrument (not the whole portfolio).
class PositionTracker(object):
    def __init__(self, round_quantity):
        self.round_quantity = round_quantity
        self.reset()

    def reset(self):
        self.__pnl = 0.0
        self.__avgPrice = 0.0  # Volume weighted average price per share.
        self.__position = 0.0
        self.__commissions = 0.0
        self.__totalCommited = 0.0  # The total amount commited to this position.

    def getPosition(self):
        return self.__position

    def getAvgPrice(self):
        return self.__avgPrice

    def getCommissions(self):
        return self.__commissions

    def getPnL(self, price=None, includeCommissions=True):
        """
        Return the PnL that would result if closing the position a the given price.
        Note that this will be different if commissions are used when the trade is executed.
        """

        ret = self.__pnl
        if price:
            ret += (price - self.__avgPrice) * self.__position
        if includeCommissions:
            ret -= self.__commissions
        return ret

    def getReturn(self, price=None, includeCommissions=True):
        ret = 0
        pnl = self.getPnL(price=price, includeCommissions=includeCommissions)
        if self.__totalCommited != 0:
            ret = pnl / float(self.__totalCommited)
        return ret

    def __openNewPosition(self, quantity, price):
        self.__avgPrice = price
        self.__position = quantity
        self.__totalCommited = self.__avgPrice * abs(self.__position)

    def __extendCurrentPosition(self, quantity, price):
        newPosition = self.round_quantity(self.__position + quantity)
        self.__avgPrice = (self.__avgPrice * abs(self.__position) + price * abs(quantity)) / abs(float(newPosition))
        self.__position = newPosition
        self.__totalCommited = self.__avgPrice * abs(self.__position)

    def __reduceCurrentPosition(self, quantity, price):
        # Check that we're closing or reducing partially
        assert self.round_quantity(abs(self.__position) - abs(quantity)) >= 0
        pnl = (price - self.__avgPrice) * quantity * -1

        self.__pnl += pnl
        self.__position = self.round_quantity(self.__position + quantity)
        if self.__position == 0:
            self.__avgPrice = 0.0

    def update(self, quantity, price, commission):
        assert quantity != 0, "Invalid quantity"
        assert price > 0, "Invalid price"
        assert commission >= 0, "Invalid commission"

        if self.__position == 0:
            self.__openNewPosition(quantity, price)
        else:
            # Are we extending the current position or going in the opposite direction ?
            currPosDirection = math.copysign(1, self.__position)
            tradeDirection = math.copysign(1, quantity)

            if currPosDirection == tradeDirection:
                self.__extendCurrentPosition(quantity, price)
            else:
                # If we're going in the opposite direction we could be:
                # 1: Partially reducing the current position.
                # 2: Completely closing the current position.
                # 3: Completely closing the current position and opening a new one in the opposite direction.
                if abs(quantity) <= abs(self.__position):
                    self.__reduceCurrentPosition(quantity, price)
                else:
                    newPos = self.__position + quantity
                    self.__reduceCurrentPosition(self.__position * -1, price)
                    self.__openNewPosition(newPos, price)

        self.__commissions += commission

    def buy(self, quantity, price, commission=0.0):
        assert quantity > 0, "Invalid quantity"
        self.update(quantity, price, commission)

    def sell(self, quantity, price, commission=0.0):
        assert quantity > 0, "Invalid quantity"
        self.update(quantity * -1, price, commission)



from datetime import datetime


class Execution:
    def __init__(self, price: float, quantity: float, commission: float, date5ime: datetime):
        self.__price = price
        self.__quantity = quantity
        self.__commission = commission
        self.__dateTime = date5ime

    def __str__(self):
        return "%s - Price: %s - Amount: %s - Fee: %s" % (
            self.__dateTime, self.__price, self.__quantity, self.__commission)

    @property
    def price(self):
        """Returns the fill price."""
        return self.__price

    @property
    def quantity(self):
        """Returns the quantity."""
        return self.__quantity

    @property
    def total(self):
        return self.__quantity * self.__price

    @property
    def commission(self):
        """Returns the commission applied."""
        return self.__commission

    @property
    def datetime(self):
        """Returns the :class:`datetime.datetime` when the order was executed."""
        return self.__dateTime



import abc

from myalgo import logger
from myalgo.broker import BackTestBroker
from myalgo.event import Event, Dispatcher
from myalgo.order import Action, LimitOrder, Order
from myalgo.strategy import position


class BaseStrategy:
    LOGGER_NAME = "BaseStrategyLog"

    def __init__(self, broker: BackTestBroker):
        self.__broker = broker
        self.__activePositions = set()
        self.__orderToPosition = {}
        self.__barsProcessedEvent = Event()
        self.__analyzers = []
        self.__namedAnalyzers = {}
        self.__resampledBarFeeds = []
        self.__dispatcher = Dispatcher()
        self.__broker.order_events.subscribe(self.__onOrderEvent)
        self.bar_feed.bar_events.subscribe(self.__onBars)

        # onStart will be called once all subjects are started.
        self.__dispatcher.getStartEvent().subscribe(self.onStart)
        self.__dispatcher.getIdleEvent().subscribe(self.__onIdle)

        # It is important to dispatch broker events before feed events, specially if we're backtesting.
        self.__dispatcher.addSubject(self.broker)
        self.__dispatcher.addSubject(self.bar_feed)

        # Initialize logging.
        self.__logger = logger.get_logger(BaseStrategy.LOGGER_NAME)

    @property
    def use_event_datetime_logs(self):
        return logger.Formatter.DATETIME_HOOK is self.dispatcher.getCurrentDateTime

    @use_event_datetime_logs.setter
    def use_event_datetime_logs(self, use: bool):
        if use:
            logger.Formatter.DATETIME_HOOK = self.dispatcher.getCurrentDateTime
        else:
            logger.Formatter.DATETIME_HOOK = None

    @property
    def broker(self):
        return self.__broker

    @property
    def bars_processed_events(self):
        return self.__barsProcessedEvent

    @property
    def bar_feed(self):
        return self.broker.bar_feed

    @property
    def feed(self):
        return self.bar_feed

    @property
    def current_datetime(self):
        return self.bar_feed.current_datetime

    @property
    def dispatcher(self):
        return self.__dispatcher

    def last_price(self, instrument: str):
        ret = None
        bar = self.bar_feed.last_bar(instrument)
        if bar is not None:
            ret = bar.price
        return ret

    def limitOrder(self, instrument: str, limit_price: float, quantity: float, goodTill_canceled: bool = False,
                   all_or_none: bool = False):
        ret = None
        if quantity > 0:
            ret: LimitOrder = self.broker.create_limit_order(Action.BUY, instrument, limit_price,
                                                             quantity)
        elif quantity < 0:
            ret: LimitOrder = self.broker.create_limit_order(Action.SELL, instrument, limit_price,
                                                             quantity * -1)
        if ret:
            ret.good_till_canceled = goodTill_canceled
            ret.all_or_one = all_or_none
            self.broker.submit_order(ret)
        return ret

    def stopLimitOrder(self, instrument, stopPrice, limitPrice, quantity, goodTillCanceled=False, allOrNone=False):
        """Submits a stop limit order.

        :param instrument: Instrument identifier.
        :type instrument: string.
        :param stopPrice: Stop price.
        :type stopPrice: float.
        :param limitPrice: Limit price.
        :type limitPrice: float.
        :param quantity: The amount of shares. Positive means buy, negative means sell.
        :type quantity: int/float.
        :param goodTillCanceled: True if the order is good till canceled. If False then the order gets automatically canceled when the session closes.
        :type goodTillCanceled: boolean.
        :param allOrNone: True if the order should be completely filled or not at all.
        :type allOrNone: boolean.
        :rtype: The :class:`pyalgotrade.broker.StopLimitOrder` submitted.
        """

        ret = None
        if quantity > 0:
            ret = self.broker.create_stop_limit_order(Action.BUY, instrument, stopPrice, limitPrice, quantity)
        elif quantity < 0:
            ret = self.broker.create_stop_limit_order(Action.SELL, instrument, stopPrice, limitPrice, quantity * -1)
        if ret:
            ret.good_till_canceled = goodTillCanceled
            ret.all_or_one = allOrNone
            self.broker.submit_order(ret)
        return ret

    def enterLong(self, instrument, quantity, goodTillCanceled=False, allOrNone=False):
        """Generates a buy :class:`pyalgotrade.broker.MarketOrder` to enter a long position.

        :param instrument: Instrument identifier.
        :type instrument: string.
        :param quantity: Entry order quantity.
        :type quantity: int.
        :param goodTillCanceled: True if the entry order is good till canceled. If False then the order gets automatically canceled when the session closes.
        :type goodTillCanceled: boolean.
        :param allOrNone: True if the orders should be completely filled or not at all.
        :type allOrNone: boolean.
        :rtype: The :class:`pyalgotrade.strategy.position.Position` entered.
        """

        return position.LongPosition(self, instrument, None, None, quantity, goodTillCanceled,
                                     allOrNone)

    def enterShort(self, instrument, quantity, goodTillCanceled=False, allOrNone=False):
        """Generates a sell short :class:`pyalgotrade.broker.MarketOrder` to enter a short position.

        :param instrument: Instrument identifier.
        :type instrument: string.
        :param quantity: Entry order quantity.
        :type quantity: int.
        :param goodTillCanceled: True if the entry order is good till canceled. If False then the order gets automatically canceled when the session closes.
        :type goodTillCanceled: boolean.
        :param allOrNone: True if the orders should be completely filled or not at all.
        :type allOrNone: boolean.
        :rtype: The :class:`pyalgotrade.strategy.position.Position` entered.
        """

        return position.ShortPosition(self, instrument, None, None, quantity, goodTillCanceled,
                                      allOrNone)

    def enterLongLimit(self, instrument, limitPrice, quantity, goodTillCanceled=False, allOrNone=False):
        """Generates a buy :class:`pyalgotrade.broker.LimitOrder` to enter a long position.

        :param instrument: Instrument identifier.
        :type instrument: string.
        :param limitPrice: Limit price.
        :type limitPrice: float.
        :param quantity: Entry order quantity.
        :type quantity: int.
        :param goodTillCanceled: True if the entry order is good till canceled. If False then the order gets automatically canceled when the session closes.
        :type goodTillCanceled: boolean.
        :param allOrNone: True if the orders should be completely filled or not at all.
        :type allOrNone: boolean.
        :rtype: The :class:`pyalgotrade.strategy.position.Position` entered.
        """

        return position.LongPosition(self, instrument, None, limitPrice, quantity,
                                     goodTillCanceled, allOrNone)

    def enterShortLimit(self, instrument, limitPrice, quantity, goodTillCanceled=False, allOrNone=False):
        """Generates a sell short :class:`pyalgotrade.broker.LimitOrder` to enter a short position.

        :param instrument: Instrument identifier.
        :type instrument: string.
        :param limitPrice: Limit price.
        :type limitPrice: float.
        :param quantity: Entry order quantity.
        :type quantity: int.
        :param goodTillCanceled: True if the entry order is good till canceled. If False then the order gets automatically canceled when the session closes.
        :type goodTillCanceled: boolean.
        :param allOrNone: True if the orders should be completely filled or not at all.
        :type allOrNone: boolean.
        :rtype: The :class:`pyalgotrade.strategy.position.Position` entered.
        """

        return position.ShortPosition(self, instrument, None, limitPrice, quantity,
                                      goodTillCanceled, allOrNone)

    def enterLongStop(self, instrument, stopPrice, quantity, goodTillCanceled=False, allOrNone=False):
        """Generates a buy :class:`pyalgotrade.broker.StopOrder` to enter a long position.

        :param instrument: Instrument identifier.
        :type instrument: string.
        :param stopPrice: Stop price.
        :type stopPrice: float.
        :param quantity: Entry order quantity.
        :type quantity: int.
        :param goodTillCanceled: True if the entry order is good till canceled. If False then the order gets automatically canceled when the session closes.
        :type goodTillCanceled: boolean.
        :param allOrNone: True if the orders should be completely filled or not at all.
        :type allOrNone: boolean.
        :rtype: The :class:`pyalgotrade.strategy.position.Position` entered.
        """

        return position.LongPosition(self, instrument, stopPrice, None, quantity, goodTillCanceled,
                                     allOrNone)

    def enterShortStop(self, instrument, stopPrice, quantity, goodTillCanceled=False, allOrNone=False):
        """Generates a sell short :class:`pyalgotrade.broker.StopOrder` to enter a short position.

        :param instrument: Instrument identifier.
        :type instrument: string.
        :param stopPrice: Stop price.
        :type stopPrice: float.
        :param quantity: Entry order quantity.
        :type quantity: int.
        :param goodTillCanceled: True if the entry order is good till canceled. If False then the order gets automatically canceled when the session closes.
        :type goodTillCanceled: boolean.
        :param allOrNone: True if the orders should be completely filled or not at all.
        :type allOrNone: boolean.
        :rtype: The :class:`pyalgotrade.strategy.position.Position` entered.
        """

        return position.ShortPosition(self, instrument, stopPrice, None, quantity,
                                      goodTillCanceled, allOrNone)

    def enterLongStopLimit(self, instrument, stopPrice, limitPrice, quantity, goodTillCanceled=False, allOrNone=False):
        """Generates a buy :class:`pyalgotrade.broker.StopLimitOrder` order to enter a long position.

        :param instrument: Instrument identifier.
        :type instrument: string.
        :param stopPrice: Stop price.
        :type stopPrice: float.
        :param limitPrice: Limit price.
        :type limitPrice: float.
        :param quantity: Entry order quantity.
        :type quantity: int.
        :param goodTillCanceled: True if the entry order is good till canceled. If False then the order gets automatically canceled when the session closes.
        :type goodTillCanceled: boolean.
        :param allOrNone: True if the orders should be completely filled or not at all.
        :type allOrNone: boolean.
        :rtype: The :class:`pyalgotrade.strategy.position.Position` entered.
        """

        return position.LongPosition(self, instrument, stopPrice, limitPrice, quantity,
                                     goodTillCanceled, allOrNone)

    def enterShortStopLimit(self, instrument, stopPrice, limitPrice, quantity, goodTillCanceled=False, allOrNone=False):
        """Generates a sell short :class:`pyalgotrade.broker.StopLimitOrder` order to enter a short position.

        :param instrument: Instrument identifier.
        :type instrument: string.
        :param stopPrice: The Stop price.
        :type stopPrice: float.
        :param limitPrice: Limit price.
        :type limitPrice: float.
        :param quantity: Entry order quantity.
        :type quantity: int.
        :param goodTillCanceled: True if the entry order is good till canceled. If False then the order gets automatically canceled when the session closes.
        :type goodTillCanceled: boolean.
        :param allOrNone: True if the orders should be completely filled or not at all.
        :type allOrNone: boolean.
        :rtype: The :class:`pyalgotrade.strategy.position.Position` entered.
        """

        return position.ShortPosition(self, instrument, stopPrice, limitPrice, quantity,
                                      goodTillCanceled, allOrNone)

    @property
    def result(self):
        return self.broker.equity

    @property
    def active_positions(self):
        return self.__activePositions

    @property
    def order_to_positions(self):
        return self.order_to_positions

    @property
    def logger(self):
        return self.__logger

    def debug(self, msg):
        """Logs a message with level DEBUG on the strategy logger."""
        self.logger.debug(msg)

    def info(self, msg):
        """Logs a message with level INFO on the strategy logger."""
        self.logger.info(msg)

    def warning(self, msg):
        """Logs a message with level WARNING on the strategy logger."""
        self.logger.warning(msg)

    def error(self, msg):
        """Logs a message with level ERROR on the strategy logger."""
        self.logger.error(msg)

    def critical(self, msg):
        """Logs a message with level CRITICAL on the strategy logger."""
        self.logger.critical(msg)

    def registerPositionOrder(self, position, order: Order):
        self.__activePositions.add(position)
        assert order.is_active  # Why register an inactive order ?
        self.__orderToPosition[order.id] = position

    def unregisterPositionOrder(self, position, order: Order):
        del self.__orderToPosition[order.id]

    def unregisterPosition(self, position):
        assert (not position.isOpen())
        self.__activePositions.remove(position)

    def __notifyAnalyzers(self, lambdaExpression):
        for s in self.__analyzers:
            lambdaExpression(s)

    def attachAnalyzerEx(self, strategyAnalyzer, name=None):
        if strategyAnalyzer not in self.__analyzers:
            if name is not None:
                if name in self.__namedAnalyzers:
                    raise Exception("A different analyzer named '%s' was already attached" % name)
                self.__namedAnalyzers[name] = strategyAnalyzer

            strategyAnalyzer.beforeAttach(self)
            self.__analyzers.append(strategyAnalyzer)
            strategyAnalyzer.attached(self)

    """
        下面开始是事件
    """

    def onEnterOk(self, position):
        """Override (optional) to get notified when the order submitted to enter a position was filled. The default implementation is empty.

        :param position: A position returned by any of the enterLongXXX or enterShortXXX methods.
        :type position: :class:`pyalgotrade.strategy.position.Position`.
        """
        pass

    def onEnterCanceled(self, position):
        """Override (optional) to get notified when the order submitted to enter a position was canceled. The default implementation is empty.

        :param position: A position returned by any of the enterLongXXX or enterShortXXX methods.
        :type position: :class:`pyalgotrade.strategy.position.Position`.
        """
        pass

    # Called when the exit order for a position was filled.
    def onExitOk(self, position):
        """Override (optional) to get notified when the order submitted to exit a position was filled. The default implementation is empty.

        :param position: A position returned by any of the enterLongXXX or enterShortXXX methods.
        :type position: :class:`pyalgotrade.strategy.position.Position`.
        """
        pass

    # Called when the exit order for a position was canceled.
    def onExitCanceled(self, position):
        """Override (optional) to get notified when the order submitted to exit a position was canceled. The default implementation is empty.

        :param position: A position returned by any of the enterLongXXX or enterShortXXX methods.
        :type position: :class:`pyalgotrade.strategy.position.Position`.
        """
        pass

    """Base class for strategies. """

    def onStart(self):
        """Override (optional) to get notified when the strategy starts executing. The default implementation is empty. """
        pass

    def onFinish(self, bars):
        """Override (optional) to get notified when the strategy finished executing. The default implementation is empty.

        :param bars: The last bars processed.
        :type bars: :class:`pyalgotrade.bar.Bars`.
        """
        pass

    def onIdle(self):
        """Override (optional) to get notified when there are no events.

       .. note::
            In a pure backtesting scenario this will not be called.
        """
        pass

    @abc.abstractmethod
    def onBars(self, datetime, bars):
        """Override (**mandatory**) to get notified when new bars are available. The default implementation raises an Exception.

        **This is the method to override to enter your trading logic and enter/exit positions**.

        :param bars: The current bars.
        :type bars: :class:`pyalgotrade.bar.Bars`.
        """
        raise NotImplementedError()

    def onOrderUpdated(self, order):
        """Override (optional) to get notified when an order gets updated.

        :param order: The order updated.
        :type order: :class:`pyalgotrade.broker.Order`.
        """
        pass

    def __onIdle(self):
        # Force a resample check to avoid depending solely on the underlying
        # barfeed events.
        for resampledBarFeed in self.__resampledBarFeeds:
            resampledBarFeed.checkNow(self.current_datetime)

        self.onIdle()

    def __onOrderEvent(self, broker_, orderEvent):
        order = orderEvent.order
        self.onOrderUpdated(order)

        # Notify the position about the order event.
        pos = self.__orderToPosition.get(order.id, None)
        if pos is not None:
            # Unlink the order from the position if its not active anymore.
            if not order.is_active:
                self.unregisterPositionOrder(pos, order)

            pos.onOrderEvent(orderEvent)

    def __onBars(self, dateTime, bars1, bars2):
        # THE ORDER HERE IS VERY IMPORTANT

        # 1: Let analyzers process bars.
        self.__notifyAnalyzers(lambda s: s.beforeOnBars(self, bars2))
        # 2: Let the strategy process current bars and submit orders.
        self.onBars(dateTime, bars2)

        # 3: Notify that the bars were processed.
        self.__barsProcessedEvent.emit(self, bars2)

    def run(self):
        """Call once (**and only once**) to run the strategy."""
        self.__dispatcher.run()

        if self.bar_feed.last_bars is not None:
            self.onFinish(self.bar_feed.last_bars)
        else:
            raise Exception("Feed was empty")

    def stop(self):
        """Stops a running strategy."""
        self.__dispatcher.stop()

    def attachAnalyzer(self, strategyAnalyzer):
        self.attachAnalyzerEx(strategyAnalyzer)

    def getNamedAnalyzer(self, name):
        return self.__namedAnalyzers.get(name, None)



import myalgo.broker as broker
import myalgo.bar as bar
import myalgo.event as event
import myalgo.feed as feed
import myalgo.order as order
import myalgo.strategy as strategy
import myalgo.stratanalyzer as stratanalyzer


class Event(object):
    def __init__(self):
        self.__handlers = []
        self.__deferred = []
        self.__emitting = 0

    def __subscribe_impl(self, handler):
        assert not self.__emitting
        if handler not in self.__handlers:
            self.__handlers.append(handler)

    def __unsubscribe_impl(self, handler):
        assert not self.__emitting
        self.__handlers.remove(handler)

    def __apply_changes(self):
        assert not self.__emitting
        for action, param in self.__deferred:
            action(param)
        self.__deferred = []

    def subscribe(self, handler):
        if self.__emitting:
            self.__deferred.append((self.__subscribe_impl, handler))
        elif handler not in self.__handlers:
            self.__subscribe_impl(handler)

    def unsubscribe(self, handler):
        if self.__emitting:
            self.__deferred.append((self.__unsubscribe_impl, handler))
        else:
            self.__unsubscribe_impl(handler)

    def emit(self, *args, **kwargs):
        try:
            self.__emitting += 1
            for handler in self.__handlers:
                handler(*args, **kwargs)
        finally:
            self.__emitting -= 1
            if not self.__emitting:
                self.__apply_changes()




from myalgo.bar.bars import Bars
from myalgo.event import Event, Subject


class BarFeed(Subject):
    def __init__(self):
        self.__bars = []

        self.__bar_events = Event()
        self.__feed_reset_event = Event()

        self.__current_bar_index = 0

        self.__started = False

        super(BarFeed, self).__init__()

    @property
    def bar_events(self):
        return self.__bar_events

    @property
    def pos(self):
        return self.__current_bar_index

    @property
    def bars(self):
        return self.__bars

    @property
    def current_bars(self):
        """

        :return:
        """
        if self.__current_bar_index >= len(self.bars):
            return None

        return self.bars[self.__current_bar_index]

    @property
    def next_bars(self):
        self.__current_bar_index += 1
        now = self.current_bars
        return now

    @property
    def current_datetime(self):
        return self.current_bars.datetime if self.current_bars else None

    @property
    def last_bars(self):
        return self.bars[self.pos - 1] if self.pos > 0 and self.__current_bar_index < len(self.bars) + 1 else None

    def last_bar(self, instrument):
        return self.last_bars.bar(instrument) if self.last_bars is not None else None

    def start(self):
        super(BarFeed, self).start()
        self.__started = True

    def stop(self):
        pass

    def join(self):
        pass

    def eof(self):
        return self.__current_bar_index >= len(self.bars)

    def peek_datetime(self):
        return self.current_datetime

    def dispatch(self):
        current_bars = self.next_bars
        if current_bars is not None:
            self.__bar_events.emit(current_bars.datetime, self.last_bars, self.current_bars)
        return current_bars is not None

    @bars.setter
    def bars(self, value):
        self.__bars = [Bars(bar_dict=i) for i in value]
        self.__feed_reset_event.emit(self.__bars)

    @property
    def instruments(self):
        """
            暂时我们只认第一个
        :return:
        """
        return self.__bars[0].instruments if len(self.__bars) > 0 else []

    @property
    def feed_reset_event(self):
        return self.__feed_reset_event



import logging
import threading

initLock = threading.Lock()
rootLoggerInitialized = False

log_format = "%(asctime)s %(name)s [%(levelname)s] %(message)s"
level = logging.DEBUG
file_log = None  # File name
console_log = True


def init_handler(handler):
    handler.setFormatter(Formatter(log_format))


def init_logger(logger):
    logger.setLevel(level)

    if file_log is not None:
        file_handler = logging.FileHandler(file_log)
        init_handler(file_handler)
        logger.addHandler(file_handler)

    if console_log:
        console_handler = logging.StreamHandler()
        init_handler(console_handler)
        logger.addHandler(console_handler)


def initialize():
    global rootLoggerInitialized
    with initLock:
        if not rootLoggerInitialized:
            init_logger(logging.getLogger())
            rootLoggerInitialized = True


def get_logger(name=None):
    initialize()
    return logging.getLogger(name)


# This formatter provides a way to hook in formatTime.
class Formatter(logging.Formatter):
    DATETIME_HOOK = None

    def formatTime(self, record, date_fmt=None):
        new_datetime = None

        if Formatter.DATETIME_HOOK is not None:
            new_datetime = Formatter.DATETIME_HOOK()

        if new_datetime is None:
            ret = super(Formatter, self).formatTime(record, date_fmt)
        else:
            ret = str(new_datetime)
        return ret



from myalgo.order.order import Order, State


class OrderEvent:
    def __init__(self, order: Order, _type: State, _info: str or None):
        self.__order = order
        self.__type = _type
        self.__info = _info

    @property
    def info(self):
        return self.__info

    @property
    def type(self):
        return self.__type

    @property
    def order(self):
        return self.__order



def get_change_percentage(actual, prev):
    if actual is None or prev is None or prev == 0:
        raise Exception("Invalid values")

    diff = actual - prev
    ret = diff / float(abs(prev))
    return ret


def safe_min(left, right):
    if left is None:
        return right
    elif right is None:
        return left
    else:
        return min(left, right)


def safe_max(left, right):
    if left is None:
        return right
    elif right is None:
        return left
    else:
        return max(left, right)



import six


class Bars(object):
    """A group of :class:`Bar` objects.

    :param barDict: A map of instrument to :class:`Bar` objects.
    :type barDict: map.

    .. note::
        All bars must have the same datetime.
    """

    def __init__(self, bar_dict):
        if len(bar_dict) == 0:
            raise Exception("No bars supplied")

        # Check that bar datetimes are in sync
        first_date_time = None
        first_instrument = None
        for instrument, currentBar in six.iteritems(bar_dict):
            if first_date_time is None:
                first_date_time = currentBar.start_date
                first_instrument = instrument
            elif currentBar.start_date != first_date_time:
                raise Exception("Bar data times are not in sync. %s %s != %s %s" % (
                    instrument,
                    currentBar.start_date,
                    first_instrument,
                    first_date_time
                ))

        self.__barDict = bar_dict
        self.__dateTime = first_date_time

    def __getitem__(self, instrument):
        return self.__barDict[instrument]

    def __contains__(self, instrument):
        return instrument in self.__barDict

    @property
    def items(self):
        return list(self.__barDict.items())

    @property
    def keys(self):
        return list(self.__barDict.keys())

    @property
    def instruments(self):
        """Returns the instrument symbols."""
        return list(self.__barDict.keys())

    @property
    def datetime(self):
        """Returns the :class:`datetime.datetime` for this set of bars."""
        return self.__dateTime

    def bar(self, instrument):
        return self.__barDict.get(instrument, None)



import abc
import functools
from datetime import datetime

from myalgo.bar import Bar
from myalgo.order import fill
from myalgo.order.action import Action
from myalgo.order.execution import Execution
from myalgo.order.state import State
from myalgo.order.type import Type


class Order:
    def __init__(self, type_: Type, action: Action, instrument, quantity: float, round_quantity=int):
        assert quantity is not None and quantity > 0
        assert type_ in Type

        self.__state = State.INITIAL
        self.__type = type_
        self.__action = action
        self.__quantity = quantity
        self.__roundQuantity = round_quantity
        self.__executionInfo = []
        self.__instrument = instrument
        self.__good_till_canceled = False
        self.__allOrNone = False
        self.__id = None
        self.__submitted_at = None
        self.__canceled_at = None
        self.__accepted_at = None

    def __repr__(self):
        return f'{self.instrument} {self.type} {self.action} {self.state} QUANT:{self.quantity}'

    @abc.abstractmethod
    def process(self, broker, bar1: Bar, bar2: Bar):
        return NotImplementedError('Not implemented!')

    @property
    def round_quantity(self):
        return self.__roundQuantity

    @property
    def good_till_canceled(self):
        return self.__good_till_canceled

    @good_till_canceled.setter
    def good_till_canceled(self, value):
        assert self.__state == State.INITIAL
        self.__good_till_canceled = value

    @property
    def all_or_one(self):
        return self.__allOrNone

    @all_or_one.setter
    def all_or_one(self, value):
        assert self.__state == State.INITIAL
        self.__allOrNone = value

    @property
    def type(self):
        return self.__type

    @property
    def is_buy(self):
        return self.action in [Action.BUY, Action.BUY_TO_COVER]

    @property
    def is_sell(self):
        return self.action in [Action.SELL, Action.SELL_SHORT]

    @property
    def is_active(self):
        return self.state not in [State.CANCELED, State.FILLED]

    @property
    def action(self):
        return self.__action

    @property
    def instrument(self):
        return self.__instrument

    @property
    def quantity(self):
        return self.__quantity

    def append_execution(self, execution: Execution):
        self.__executionInfo.append(execution)

    @property
    def executions(self):
        return self.__executionInfo

    @property
    def filled(self):
        def sum_quantity(x, y):
            return x + y.quantity

        return functools.reduce(sum_quantity, self.executions, 0)

    @property
    def filled_cost(self):
        def sum_total(x, y):
            return x + y.total

        return functools.reduce(sum_total, self.executions, 0)

    @property
    def total_commission(self):
        def sum_total(x, y: Execution):
            return x + y.commission

        return functools.reduce(sum_total, self.executions, 0)

    @property
    def avg_fill_price(self):
        assert self.filled , 'order not filled at all!'
        return self.filled_cost / float(self.filled)

    @property
    def remain(self):
        return self.__roundQuantity(self.quantity - self.filled)

    @property
    def id(self):
        return self.__id

    def submitted(self, _id, at: datetime):
        self.__id = _id
        self.__submitted_at = at
        return self

    def canceled(self, at: datetime):
        self.__canceled_at = at
        return self

    def accepted(self, at: datetime):
        self.__accepted_at = at
        return self

    @property
    def submitted_at(self):
        return self.__submitted_at

    @property
    def canceled_at(self):
        return self.__canceled_at

    @property
    def accepted_at(self):
        return self.__accepted_at

    @property
    def state(self):
        if self.canceled_at is not None:
            return State.CANCELED
        if self.submitted_at is None:
            return State.INITIAL

        if self.accepted_at is None:
            return State.SUBMITTED

        if self.filled is 0:
            return State.ACCEPTED
        elif self.remain is 0:
            return State.FILLED
        else:
            return State.PARTIALLY_FILLED

    @property
    def is_canceled(self):
        return self.canceled_at is not None

    @property
    def is_submitted(self):
        return self.state == State.SUBMITTED

    @property
    def is_accepted(self):
        return self.state == State.ACCEPTED

    @property
    def is_filled(self):
        return self.state == State.FILLED

    @property
    def is_partially_filled(self):
        return self.state == State.PARTIALLY_FILLED

    @property
    def is_initial(self):
        return self.state == State.INITIAL

    @property
    def finish_datetime(self):
        if self.executions is not None and len(self.executions) > 0:
            return self.executions[-1].datetime
        return None


class MarketOrder(Order):
    """
        Base Class for market orders
    """

    def __init__(self, action, instrument, quantity, on_close, round_quantity):
        super(MarketOrder, self).__init__(Type.MARKET, action, instrument, quantity, round_quantity)
        self.__onClose = on_close

    @property
    def fill_on_close(self):
        return self.__onClose

    def process(self, broker, bar1: Bar, bar2: Bar):
        if self.fill_on_close:
            price = bar2.ask_close if self.is_buy else bar2.bid_close
        else:
            price = bar2.ask_open if self.is_buy else bar2.bid_open
        assert price is not None
        return fill.FillInfo(price, self.quantity)


class LimitOrder(Order):
    def __init__(self, action: Action, instrument: str, limit_price: float, quantity: float, round_quantity=int):
        super(LimitOrder, self).__init__(Type.LIMIT, action, instrument, quantity, round_quantity)
        self.__limitPrice = limit_price

    @property
    def price(self):
        """Returns the limit price."""
        return self.__limitPrice

    def __repr__(self):
        return super(LimitOrder, self).__repr__()+f'\t PRICE:{self.price}'
    def process(self, broker, bar1: Bar, bar2: Bar):
        price = fill.get_limit_price_trigger(self.action, self.price, bar1, bar2)
        return None if price is None else fill.FillInfo(price, self.quantity)


class StopOrder(Order):
    """Base class for stop orders.

    .. note::

        This is a base class and should not be used directly.
    """

    def __init__(self, action, instrument, price, quantity, round_quantity):
        super(StopOrder, self).__init__(Type.STOP, action, instrument, quantity, round_quantity)

        self.__stopPrice = price
        self.__stopHit = False

    @property
    def stop_hit(self):
        return self.__stopHit

    @property
    def price(self):
        """Returns the stop price."""
        return self.__stopPrice

    def process(self, broker, bar1: Bar, bar2: Bar):
        price_trigger = None
        if not self.__stopHit:
            price_trigger = fill.get_stop_price_trigger(self.action, self.price, bar1, bar2)
        if price_trigger is not None:
            self.__stopHit = True
        if self.stop_hit:
            """
                由于往往缺少数据，这里直接认为所有量全成交
            """
            fill_size = self.quantity
            if price_trigger is not None:
                price = price_trigger
            else:
                price = bar2.ask_open if self.is_buy else bar2.bid_open
            return fill.FillInfo(price, fill_size)
        return None


class StopLimitOrder(Order):
    """Base class for stop limit orders.

    .. note::

        This is a base class and should not be used directly.
    """

    def __init__(self, action, instrument, stop_price, limit_price, quantity, round_quantity):
        super(StopLimitOrder, self).__init__(Type.STOP_LIMIT, action, instrument, quantity, round_quantity)

        self.__stopPrice = stop_price
        self.__limitPrice = limit_price
        self.__stopHit = False

    @property
    def stop_hit(self):
        return self.__stopHit

    @property
    def stop_price(self):
        """Returns the stop price."""
        return self.__stopPrice

    @property
    def limit_price(self):
        """Returns the limit price."""
        return self.__limitPrice

    def process(self, broker, bar1: Bar, bar2: Bar):
        price_trigger = None
        if not self.__stopHit:
            price_trigger = fill.get_stop_price_trigger(self.action, self.stop_price, bar1, bar2)

        self.__stopHit = price_trigger is not None

        if self.stop_hit:
            fill_size = self.quantity

            price = fill.get_limit_price_trigger(
                self.action,
                self.limit_price,
                bar1,
                bar2
            )
            if price is not None:
                # If we just hit the stop price, we need to make additional checks.
                if price_trigger is not None:
                    if self.is_buy:
                        # If the stop price triggered is lower than the limit price, then use that one.
                        # Else use the limit price.
                        price = min(price_trigger, self.limit_price)
                    else:
                        # If the stop price triggered is greater than the limit price, then use that one.
                        # Else use the limit price.
                        price = max(price_trigger, self.limit_price)
                return fill.FillInfo(price, fill_size)
        return None



from datetime import datetime

import six

import myalgo.logger as logger
from myalgo.bar import Bar, Bars
from myalgo.broker.commission import Commission
from myalgo.config import dispatchprio
from myalgo.event import Event
from myalgo.event import Subject
from myalgo.feed.barfeed import BarFeed
from myalgo.order import LimitOrder, Action, Order, Execution, OrderEvent, State, MarketOrder, StopLimitOrder, StopOrder
from myalgo.order.fill import FillInfo


class BackTestBroker(Subject):
    LOGGER_NAME = "back_test_log"

    def __init__(self, cash: float, bar_feed: BarFeed, commission: Commission, round_quantity=lambda x: int(x)):

        super(BackTestBroker, self).__init__()
        self.__commission = commission
        self.__bar_feed = bar_feed
        self.__cash = self.__initial_cash = cash

        self.__next_order_id = 0
        self.__active_orders = {}
        self.__quantities = {}
        self.__logger = logger.get_logger("Broker_log")

        self.__order_events = Event()

        self.__round = round_quantity

        self.__bar_feed.bar_events.subscribe(self.on_bars)
        self.__bar_feed.feed_reset_event.subscribe(self.__on_feed_change)
        self.__started = False

        self.instruments = bar_feed.instruments

    @property
    def instruments(self):
        return self.__instruments

    @property
    def quantities(self):
        return self.__quantities

    @instruments.setter
    def instruments(self, lists):
        assert not self.started, 'should not change instrument when started'
        self.__instruments = lists
        for instrument in lists:
            self.__quantities[instrument] = 0

    def __reset_instruments(self):
        self.instruments = self.bar_feed.instruments


    def __on_feed_change(self, bars):
        assert not self.started, 'should not change feed when started'
        self.__reset_instruments()

    @property
    def started(self):
        return self.__started

    @property
    def dispatch_priority(self):
        return dispatchprio.BROKER

    @property
    def next_order_id(self):
        ret = self.__next_order_id
        self.__next_order_id += 1
        return ret

    @property
    def order_events(self):
        return self.__order_events

    def register_order(self, order_: Order):
        assert (order_.id not in self.__active_orders)
        assert (order_.id is not None)
        self.__active_orders[order_.id] = order_

    def unregister_order(self, order_: Order):
        assert (order_.id in self.__active_orders)
        assert (order_.id is not None)
        del self.__active_orders[order_.id]

    @property
    def logger(self):
        return self.__logger

    def get_bar(self, bars, instrument):
        ret = bars.bar(instrument)
        if ret is None:
            ret = self.__bar_feed.last_bar(instrument)
        return ret

    def cash(self, include_short=True):
        ret = self.__cash
        if not include_short and self.__bar_feed.current_bars is not None:
            bars = self.__bar_feed.current_bars
            for instrument, shares in six.iteritems(self.__quantities):
                if shares < 0:
                    instrument_price = self.get_bar(bars, instrument).in_price
                    ret += instrument_price * shares
        return ret

    @property
    def equity(self):
        """Returns the portfolio value (cash + shares * price)."""
        ret = self.cash()
        for instrument, shares in six.iteritems(self.__quantities):
            instrument_price = self.__bar_feed.last_bar(instrument).out_price
            assert instrument_price is not None, "Price for %s is missing" % instrument
            ret += instrument_price * shares
        return ret

    def create_limit_order(self, action: Action, instrument: str, limit_price: float, quantity: float):
        return LimitOrder(action, instrument, limit_price=limit_price, quantity=quantity,
                          round_quantity=self.__round)

    def create_market_order(self, action: Action, instrument: str, quantity: float, on_close=False):
        # In order to properly support market-on-close with intraday feeds I'd need to know about different
        # exchange/market trading hours and support specifying routing an order to a specific exchange/market.
        # Even if I had all this in place it would be a problem while paper-trading with a live feed since
        # I can't tell if the next bar will be the last bar of the market session or not.
        return MarketOrder(action, instrument, quantity, on_close, round_quantity=self.__round)

    def create_stop_order(self, action: Action, instrument: str, price: float, quantity: float):
        return StopOrder(action, instrument, price, quantity, round_quantity=self.__round)

    def create_stop_limit_order(self, action: Action, instrument: str, stop_price: float, limit_price: float,
                                quantity: float):
        return StopLimitOrder(action, instrument, stop_price, limit_price, quantity, round_quantity=self.__round)

    # Tries to commit an order execution.
    def commit_order_execution(self, fill_info: FillInfo, order_: Order, datetime_: datetime):
        price = fill_info.price
        quantity = fill_info.quantity

        if order_.is_buy:
            cost = price * quantity * -1
            assert (cost < 0)
            shares_delta = quantity
        elif order_.is_sell:
            cost = price * quantity
            assert (cost > 0)
            shares_delta = quantity * -1
        else:  # Unknown action
            assert False

        commission = self.__commission.calculate(order_, price, quantity)
        cost -= commission
        resulting_cash = self.__cash + cost

        # Check that we're ok on cash after the commission.
        if resulting_cash >= 0:

            # Update the order before updating internal state since addExecutionInfo may raise.
            # addExecutionInfo should switch the order state.
            execution = Execution(price, quantity, commission, datetime_)
            order_.append_execution(execution)

            # Commit the order execution.
            self.__cash = resulting_cash


            updated_shares = order_.round_quantity(
                self.__quantities[order_.instrument] + shares_delta
            )
            self.__quantities[order_.instrument] = updated_shares

            # Notify the order update
            if order_.is_filled:
                self.unregister_order(order_)
                self.notify_order_event(OrderEvent(order_, State.FILLED, execution))
            elif order_.is_partially_filled:
                self.notify_order_event(OrderEvent(order_, State.PARTIALLY_FILLED, execution))
            else:
                assert False
        else:
            self.__logger.debug("Not enough cash to fill %s order [%s] for %s share/s" % (
                order_.instrument,
                order_.id,
                order_.remain
            ))

    def submit_order(self, order_: Order):
        if order_.is_initial:
            order_.submitted(self.next_order_id, self.current_datetime)
            self.register_order(order_)
            self.notify_order_event(OrderEvent(order_, State.SUBMITTED, None))
        else:
            raise Exception("The order was already processed")

    # Return True if further processing is needed.
    def __preprocess_order(self, order_: Order, bar_: Bar):
        ret = True

        # For non-GTC orders we need to check if the order has expired.
        if not order_.good_till_canceled:

            current = bar_.start_date

            expired = current.date() > order_.accepted_at.date()

            # Cancel the order if it is expired.
            if expired:
                ret = False
                self.unregister_order(order_)
                order_.canceled(current)
                self.notify_order_event(OrderEvent(order_, State.CANCELED, "Expired"))

        return ret

    def __postprocess_order(self, order_: Order, bar_: Bar):
        # For non-GTC orders and daily (or greater) bars we need to check if orders should expire right now
        # before waiting for the next bar.
        if not order_.good_till_canceled:
            expired = False

            # Cancel the order if it will expire in the next bar.
            if expired:
                self.unregister_order(order_)
                order_.canceled(bar_.start_date)
                self.notify_order_event(OrderEvent(order_, State.CANCELLED, "Expired"))

    def __process_order(self, order_, bar1: Bar, bar2: Bar):
        if not self.__preprocess_order(order_, bar2):
            return
        # Double dispatch to the fill strategy using the concrete order type.
        fill_info = order_.process(self, bar1, bar2)
        if fill_info is not None:
            self.commit_order_execution(fill_info, order_, bar2.start_date)

        if order_.is_active:
            self.__postprocess_order(order_, bar2)

    def __on_bars_impl(self, order_, bars1: Bars, bars2: Bars):
        # IF WE'RE DEALING WITH MULTIPLE INSTRUMENTS WE SKIP ORDER PROCESSING IF THERE IS NO BAR FOR THE ORDER'S
        # INSTRUMENT TO GET THE SAME BEHAVIOUR AS IF WERE BE PROCESSING ONLY ONE INSTRUMENT.
        bar1 = bars1.bar(order_.instrument)
        bar2 = bars2.bar(order_.instrument)
        if bar1 is not None and bar2 is not None:
            # Switch from SUBMITTED -> ACCEPTED
            if order_.is_submitted:
                order_.accepted(bar2.start_date)
                self.notify_order_event(
                    OrderEvent(order_, State.ACCEPTED, None))

            if order_.is_active:
                # This may trigger orders to be added/removed from __activeOrders.
                self.__process_order(order_, bar1, bar2)
            else:
                # If an order is not active it should be because it was c11anceled in this same loop and it should
                # have been removed.
                assert order_.is_canceled
                assert order_ not in self.__active_orders

    def on_bars(self, datetime_, bars1: Bars, bars2: Bars):


        # This is to froze the orders that will be processed in this event, to avoid new getting orders introduced
        # and processed on this very same event.
        orders_to_process = list(self.__active_orders.values())



        for order_ in orders_to_process:
            # This may trigger orders to be added/removed from __activeOrders.
            self.__on_bars_impl(order_, bars1, bars2)

    @property
    def active_instruments(self):
        return [instrument for instrument, shares in six.iteritems(self.__quantities) if shares != 0]

    def active_orders(self, instrument=None):
        if instrument is None:
            ret = list(self.__active_orders.values())
        else:
            ret = [order_ for order_ in self.__active_orders.values() if order_.instrument == instrument]
        return ret

    def cancel_order(self, order_: Order):
        active_order = self.__active_orders.get(order_.id)
        if active_order is None:
            raise Exception("The order is not active anymore")
        if active_order.is_filled:
            raise Exception("Can't cancel order that has already been filled")

        self.unregister_order(active_order)
        active_order.canceled(self.current_datetime)
        self.notify_order_event(
            OrderEvent(active_order, State.CANCELED, "User requested cancellation")
        )

    @property
    def commission(self):
        """Returns the strategy used to calculate order commissions.

        :rtype: :class:`Commission`.
        """
        return self.__commission

    @commission.setter
    def commission(self, commission):
        """Sets the strategy to use to calculate order commissions.

        :param commission: An object responsible for calculating order commissions.
        :type commission: :class:`Commission`.
        """

        self.__commission = commission

    @property
    def current_datetime(self):
        return self.__bar_feed.current_datetime

    @property
    def bar_feed(self):
        return self.__bar_feed

    def start(self):
        super(BackTestBroker, self).start()
        self.__started = True

    def stop(self):
        pass

    def join(self):
        pass

    def eof(self):
        # If there are no more events in the barfeed, then there is nothing left for us to do since all processing took
        # place while processing barfeed events.
        return self.bar_feed.eof()

    def dispatch(self):
        # All events were already emitted while handling barfeed events.
        pass

    def peek_datetime(self):
        return None

    def notify_order_event(self, event: OrderEvent):
        return self.__order_events.emit(self, event)



import abc

import six

from myalgo.config import dispatchprio


@six.add_metaclass(abc.ABCMeta)
class Subject(object):

    def __init__(self):
        self.__dispatch_priority = dispatchprio.LAST

    # This may raise.
    @abc.abstractmethod
    def start(self):
        pass

    # This should not raise.
    @abc.abstractmethod
    def stop(self):
        raise NotImplementedError()

    # This should not raise.
    @abc.abstractmethod
    def join(self):
        raise NotImplementedError()

    # Return True if there are not more events to dispatch.
    @abc.abstractmethod
    def eof(self):
        raise NotImplementedError()

    # Dispatch events. If True is returned, it means that at least one event was dispatched.
    @abc.abstractmethod
    def dispatch(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def peek_datetime(self):
        # Return the datetime for the next event.
        # This is needed to properly synchronize non-realtime subjects.
        # Return None since this is a realtime subject.
        raise NotImplementedError()

    @property
    def dispatch_priority(self):
        return self.__dispatch_priority

    @dispatch_priority.setter
    def dispatch_priority(self, value):
        self.__dispatch_priority = value

    def on_dispatch_priority_registered(self, dispatcher):
        # Called when the subject is registered with a dispatcher.
        pass


from myalgo.bar import Bar
from myalgo.order import Action


class FillInfo(object):
    def __init__(self, price: float, quantity: float):
        self.__price = price
        self.__quantity = quantity

    @property
    def price(self):
        return self.__price

    @property
    def quantity(self):
        return self.__quantity


# Returns the trigger price for a Limit or StopLimit order, or None if the limit price was not yet penetrated.
def get_limit_price_trigger(action: Action, price: float, bar1: Bar, bar2: Bar):
    """
        以最稳健的方式保证除入场
    :param action:
    :param price:
    :param bar1:
    :param bar2:
    :return:
    """
    ret = None
    if action in [Action.BUY, Action.BUY_TO_COVER]:
        if bar1.ask_close > price >= bar2.ask_low:
            return price

    elif action in [Action.SELL, Action.SELL_SHORT]:
        if bar1.bid_close < price <= bar2.bid_high:
            return price

    else:  # Unknown action
        assert False
    return ret


# Returns the trigger price for a Stop or StopLimit order, or None if the stop price was not yet penetrated.
def get_stop_price_trigger(action: Action, price: float, bar1: Bar, bar2: Bar):
    ret = None

    # If the bar is above the stop price, use the open price.
    # If the bar includes the stop price, use the open price or the stop price. Whichever is better.
    if action in [Action.BUY, Action.BUY_TO_COVER]:
        if bar1.ask_close < price:
            if bar2.ask_low > price:
                ret = bar2.ask_open
            elif price <= bar2.ask_high:
                return min(bar2.ask_open, price)
    # If the bar is below the stop price, use the open price.
    # If the bar includes the stop price, use the open price or the stop price. Whichever is better.
    elif action in [Action.SELL, Action.SELL_SHORT]:
        if bar1.bid_close > price:
            if bar2.bid_high < price:
                ret = bar2.bid_open
            elif price >= bar2.bid_low:
                return max(bar2.bid_open, price)
    else:  # Unknown action
        assert False

    return ret



from myalgo.order.action import Action
from myalgo.order.event import OrderEvent
from myalgo.order.execution import Execution
from myalgo.order.order import Order, LimitOrder, StopLimitOrder, MarketOrder, StopOrder
from myalgo.order.state import State
from myalgo.order.type import Type



import datetime
import time

from sqlalchemy import Column, DateTime, String, Integer, Float, CHAR, func
from sqlalchemy import and_
from sqlalchemy import cast
from sqlalchemy import create_engine
from sqlalchemy import event
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

from myalgo.bar.bar import Bar
from myalgo.feed.barfeed import BarFeed
from myalgo.logger import get_logger

Base = declarative_base()


def make_bar_model(table_name):
    class BarModel(Base):
        __tablename__ = table_name
        id = Column(Integer, primary_key=True)
        type = Column(CHAR(45))

        ask_open = Column(Float)
        ask_low = Column(Float)
        ask_high = Column(Float)
        ask_close = Column(Float)

        bid_open = Column(Float)
        bid_low = Column(Float)
        bid_high = Column(Float)
        bid_close = Column(Float)

        start_date = Column(DateTime)
        end_date = Column(DateTime)

        def convert_to_bar(self):
            return Bar(self.start_date, self.end_date, self.ask_open, self.ask_close, self.ask_high, self.ask_low,
                       self.bid_open, self.bid_close, self.bid_high, self.bid_low, 0)

    return BarModel


class SQLiteFeed(BarFeed):
    LOGGER_NAME = "SQLITE_FEED_LOGGER"

    def __init__(self,
                 table_name='bins',
                 file_name='sqlite'):
        self.__engine = create_engine(f'sqlite:///{file_name}')
        self.__DBSession = sessionmaker(bind=self.__engine)
        self.__logger = get_logger(SQLiteFeed.LOGGER_NAME)
        self.__bar_model = make_bar_model(table_name)

        def before_cursor_execute(conn, cursor, statement,
                                  parameters, context, executemany):
            conn.info.setdefault('query_start_time', []).append(time.time())

            self.__logger.debug(f'Start Query: {statement} with params: {parameters}')

        def after_cursor_execute(conn, cursor, statement,
                                 parameters, context, executemany):
            total = time.time() - conn.info['query_start_time'].pop(-1)
            self.__logger.debug("Query Complete!")
            self.__logger.debug("Total Time: %f", total)

        event.listens_for(self.__engine, "before_cursor_execute")(before_cursor_execute)
        event.listens_for(self.__engine, "after_cursor_execute")(after_cursor_execute)

        super(SQLiteFeed, self).__init__()

        self.bars = []

    @property
    def new_session(self):
        return self.__DBSession()

    @property
    def model(self):
        return self.__bar_model

    def load_data(self, instruments, from_date, to_date):
        """将数据全部加载上来"""

        self.__logger.debug(
            f'loading data from database, for instruments: {instruments} from: {from_date} to: {to_date}')
        data_dicts = {}

        length = []
        session = self.new_session

        for instrument in instruments:
            data_dicts[instrument] = self.__fetch(session, instrument, to_date, from_date)
            length.append(len(data_dicts[instrument]))

        session.close()

        self.__logger.info('loading data from database complete!')

        m_length = min(length)

        def get_instrument(i):
            new_dicts = {}
            for inst in instruments:
                new_dicts[inst] = data_dicts[inst][i].convert_to_bar()
            return new_dicts

        result = [get_instrument(i) for i in range(m_length)]

        self.bars = result

    def __fetch(self, session, instrument, to_date=datetime.datetime(2019, 1, 1, 0, 0, 0),
                from_date=datetime.datetime(2012, 1, 1, 0, 0, 0)):
        """
            选择一段时间的交易记录，并且返回
            from_date: 开始时间
            to_date: 结束时间
            type: 种类
        """
        cursor = session.query(self.model).filter(and_(
            cast(self.model.type, String) == cast(instrument, String),
            self.model.start_date >= func.datetime(from_date),
            self.model.end_date <= func.datetime(to_date))).order_by("start_date")
        result = cursor.all()
        return result



from datetime import datetime


class Bar:

    def __init__(self, start_date: datetime, end_date: datetime, ask_open: float, ask_close: float, ask_high: float,
                 ask_low: float, bid_open: float, bid_close: float, bid_high: float,
                 bid_low, volume):
        if ask_high < ask_low:
            raise Exception("high < low on %s" % start_date)
        elif ask_high < ask_open:
            raise Exception("high < open on %s" % start_date)
        elif ask_high < ask_close:
            raise Exception("high < close on %s" % start_date)
        elif ask_low > ask_open:
            raise Exception("low > open on %s" % start_date)
        elif ask_low > ask_close:
            raise Exception("low > close on %s" % start_date)

        if bid_high < bid_low:
            raise Exception("high < low on %s" % start_date)
        elif bid_high < bid_open:
            raise Exception("high < open on %s" % start_date)
        elif bid_high < bid_close:
            raise Exception("high < close on %s" % start_date)
        elif bid_low > bid_open:
            raise Exception("low > open on %s" % start_date)
        elif bid_low > bid_close:
            raise Exception("low > close on %s" % start_date)

        self.__start_date = start_date
        self.__end_date = end_date

        self.__ask_open = ask_open
        self.__ask_close = ask_close
        self.__ask_high = ask_high
        self.__ask_low = ask_low

        self.__bid_open = bid_open
        self.__bid_close = bid_close
        self.__bid_high = bid_high
        self.__bid_low = bid_low

        self.__volume = volume

    def __getstate__(self):
        return (
            self.start_date,
            self.end_date,
            self.ask_open,
            self.ask_close,
            self.ask_high,
            self.ask_low,
            self.bid_open,
            self.bid_high,
            self.bid_low,
            self.bid_close,
            self.volume,
        )

    @property
    def start_date(self):
        return self.__start_date

    @property
    def end_date(self):
        return self.__end_date

    @property
    def ask_open(self):
        return self.__ask_open

    @property
    def ask_high(self):
        return self.__ask_high

    @property
    def ask_low(self):
        return self.__ask_low

    @property
    def ask_close(self):
        return self.__ask_close

    @property
    def bid_open(self):
        return self.__bid_open

    @property
    def bid_high(self):
        return self.__bid_high

    @property
    def bid_low(self):
        return self.__bid_low

    @property
    def bid_close(self):
        return self.__bid_close

    @property
    def volume(self):
        return self.__volume

    @property
    def out_price(self):
        return self.__bid_close

    @property
    def in_price(self):
        return self.__ask_close

    @property
    def price(self):
        return self.out_price

    @property
    def dict(self):
        return {
            "start_date":self.start_date,
            "end_date":self.end_date,
            "ask_open": self.ask_open,
            "ask_low": self.ask_low,
            "ask_high": self.ask_high,
            "ask_close": self.ask_close,
            "bid_open": self.bid_open,
            "bid_low": self.bid_low,
            "bid_high": self.bid_high,
            "bid_close": self.bid_close,
        }




from setuptools import setup,find_packages


setup(
    name='myalgo',
    version='1.2.2',
    packages=find_packages(),
    url='',
    license='',
    author='nathaniel',
    author_email='348831271@qq.com',
    description='a simple trader for CTA trading.'
)



import abc

import six

from myalgo.order.order import Order


######################################################################
# Commission models


@six.add_metaclass(abc.ABCMeta)
class Commission(object):
    """Base class for implementing different commission schemes.

    .. note::
        This is a base class and should not be used directly.
    """

    @abc.abstractmethod
    def calculate(self, order: Order, price: float, quantity: float):
        """Calculates the commission for an order execution.

        :param order: The order being executed.
        :type order: :class:`order.Order`.
        :param price: The price for each share.
        :type price: float.
        :param quantity: The order size.
        :type quantity: float.
        :rtype: float.
        """
        raise NotImplementedError()


class NoCommission(Commission):
    """A :class:`Commission` class that always returns 0."""

    def calculate(self, order: Order, price: float, quantity: float):
        return 0


class FixedPerTrade(Commission):
    """A :class:`Commission` class that charges a fixed amount for the whole trade.

    :param amount: The commission for an order.
    :type amount: float.
    """

    def __init__(self, amount):
        super(FixedPerTrade, self).__init__()
        self.__amount = amount

    def calculate(self, order: Order, price: float, quantity: float):
        ret = 0
        # Only charge the first fill.
        if order.executions is None:
            ret = self.__amount
        return ret


class TradePercentage(Commission):
    """A :class:`Commission` class that charges a percentage of the whole trade.

    :param percentage: The percentage to charge. 0.01 means 1%, and so on. It must be smaller than 1.
    :type percentage: float.
    """

    def __init__(self, percentage):
        super(TradePercentage, self).__init__()
        assert (percentage < 1)
        self.__percentage = percentage

    def calculate(self, order: Order, price: float, quantity: float):
        return price * quantity * self.__percentage
